@@ -1,7 +1,7 @@
 """Proxy for coordinates stored outside Shapely geometries
 """
 
-from shapely.geometry.base import deserialize_wkb, EMPTY
+from shapely.geometry.base import EMPTY
 from shapely.geos import lgeos
 
 
@@ -30,7 +30,8 @@ class CachingGeometryProxy(object):
         gtag = self.gtag()
         if gtag != self._gtag or self._is_empty:
             self.empty()
-            self.__geom__, n = self.factory(self.context)
+            if len(self.context) > 0:
+                self.__geom__, n = self.factory(self.context)
         self._gtag = gtag
         return self.__geom__
         
