@@ -2,28 +2,29 @@
 Proxies for libgeos, GEOS-specific exceptions, and utilities
 """
 
+import atexit
+from ctypes import (
+    CDLL, cdll, pointer, string_at, DEFAULT_MODE, c_void_p, c_size_t, c_char_p)
+from ctypes.util import find_library
+import glob
+import logging
 import os
 import re
 import sys
-import atexit
-import logging
 import threading
-from ctypes import CDLL, cdll, pointer, string_at, cast, POINTER, DEFAULT_MODE
-from ctypes import c_void_p, c_size_t, c_char_p, c_int, c_float
-from ctypes.util import find_library
+from functools import partial
 
 from .ctypes_declarations import prototype, EXCEPTION_HANDLER_FUNCTYPE
-from . import ftools
+from .errors import WKBReadingError, WKTReadingError, TopologicalError, PredicateError
 
 
 # Add message handler to this module's logger
 LOG = logging.getLogger(__name__)
-ch = logging.StreamHandler()
-LOG.addHandler(ch)
 
-if 'all' in sys.warnoptions:
-    # show GEOS messages in console with: python -W all
-    LOG.setLevel(logging.DEBUG)
+if sys.version_info[0] >= 3:
+    text_types = str
+else:
+    text_types = (str, unicode)
 
 
 # Find and load the GEOS and C libraries
@@ -37,7 +38,7 @@ def load_dll(libname, fallbacks=None, mode=DEFAULT_MODE):
             LOG.debug("Trying `CDLL(%s)`", lib)
             dll = CDLL(lib, mode=mode)
         except OSError:
-            LOG.warn("Failed `CDLL(%s)`", lib)
+            LOG.debug("Failed `CDLL(%s)`", lib)
             pass
 
     if not dll and fallbacks is not None:
@@ -47,7 +48,7 @@ def load_dll(libname, fallbacks=None, mode=DEFAULT_MODE):
                 dll = CDLL(name, mode=mode)
             except OSError:
                 # move on to the next fallback
-                LOG.warn("Failed `CDLL(%s)`", name)
+                LOG.debug("Failed `CDLL(%s)`", name)
                 pass
 
     if dll:
@@ -57,94 +58,115 @@ def load_dll(libname, fallbacks=None, mode=DEFAULT_MODE):
     else:
         # No shared library was loaded. Raise OSError.
         raise OSError(
-            "Could not find lib {0} or load any of its variants {1}.".format(
+            "Could not find lib {} or load any of its variants {}.".format(
                 libname, fallbacks or []))
 
 _lgeos = None
 
 if sys.platform.startswith('linux'):
-    _lgeos = load_dll('geos_c', fallbacks=['libgeos_c.so.1', 'libgeos_c.so'])
-    free = load_dll('c').free
+    # Test to see if we have a wheel repaired by 'auditwheel' containing its
+    # own libgeos_c
+    geos_whl_so = glob.glob(os.path.abspath(os.path.join(os.path.dirname(
+        __file__), '.libs/libgeos_c-*.so.*')))
+    if len(geos_whl_so) == 1:
+        _lgeos = CDLL(geos_whl_so[0])
+        LOG.debug("Found GEOS DLL: %r, using it.", _lgeos)
+    elif hasattr(sys, 'frozen'):
+        geos_pyinstaller_so = glob.glob(os.path.join(sys.prefix, 'libgeos_c-*.so.*'))
+        if len(geos_pyinstaller_so) == 1:
+            _lgeos = CDLL(geos_pyinstaller_so[0])
+            LOG.debug("Found GEOS DLL: %r, using it.", _lgeos)
+    elif os.getenv('CONDA_PREFIX', ''):
+        # conda package.
+        _lgeos = CDLL(os.path.join(sys.prefix, 'lib', 'libgeos_c.so'))
+    else:
+        alt_paths = [
+            'libgeos_c.so.1',
+            'libgeos_c.so',
+        ]
+        _lgeos = load_dll('geos_c', fallbacks=alt_paths)
+    # Necessary for environments with only libc.musl
+    c_alt_paths = [
+        'libc.musl-x86_64.so.1'
+    ]
+    free = load_dll('c', fallbacks=c_alt_paths).free
     free.argtypes = [c_void_p]
     free.restype = None
 
 elif sys.platform == 'darwin':
-    # First: have we already loaded GEOS through a Fiona or Rasterio?
-    # If so, let's obtain a handle to it instead of loading this module's
-    # copy to side step the mysterious issue described at
-    # https://github.com/Toblerity/Shapely/issues/324.
-    geos_mod = sys.modules.get('fiona') or sys.modules.get('rasterio')
-    if geos_mod:
-        dll_path = os.path.join(
-            os.path.dirname(geos_mod.__file__), '.dylibs',
-            'libgeos_c.1.dylib')
-        try:
-            _lgeos = CDLL(dll_path, mode=(DEFAULT_MODE | 16))
-        except OSError:
-            LOG.debug("GEOS DLL in fiona or rasterio .dylibs not found.")
+    # Test to see if we have a delocated wheel with a GEOS dylib.
+    geos_whl_dylib = os.path.abspath(os.path.join(os.path.dirname(
+        __file__), '.dylibs/libgeos_c.1.dylib'))
+
+    if os.path.exists(geos_whl_dylib):
+        handle = CDLL(None)
+        if hasattr(handle, "initGEOS_r"):
+            LOG.debug("GEOS already loaded")
+            _lgeos = handle
+        else:
+            _lgeos = CDLL(geos_whl_dylib)
+            LOG.debug("Found GEOS DLL: %r, using it.", _lgeos)
+
+    elif os.getenv('CONDA_PREFIX', ''):
+        # conda package.
+        _lgeos = CDLL(os.path.join(sys.prefix, 'lib', 'libgeos_c.dylib'))
+    else:
+        if hasattr(sys, 'frozen'):
+            try:
+                # .app file from py2app
+                alt_paths = [os.path.join(
+                    os.environ['RESOURCEPATH'], '..', 'Frameworks',
+                    'libgeos_c.dylib')]
+            except KeyError:
+                # binary from pyinstaller
+                alt_paths = [
+                    os.path.join(sys.executable, 'libgeos_c.dylib')]
+                if hasattr(sys, '_MEIPASS'):
+                    alt_paths.append(
+                        os.path.join(sys._MEIPASS, 'libgeos_c.1.dylib'))
+        else:
             alt_paths = [
                 # The Framework build from Kyng Chaos
                 "/Library/Frameworks/GEOS.framework/Versions/Current/GEOS",
                 # macports
                 '/opt/local/lib/libgeos_c.dylib',
+                # homebrew
+                '/usr/local/lib/libgeos_c.dylib',
             ]
-            _lgeos = load_dll('geos_c', fallbacks=alt_paths,
-                              mode=(DEFAULT_MODE | 16))
-        if _lgeos:
-            LOG.debug("Found %r already loaded, using it.", _lgeos)
-
-    # If neither fiona nor rasterio have been imported, or if the block
-    # above failed to assign _lgeos, we will locate and load the GEOS
-    # DLL.
-    if not _lgeos:
-        # Test to see if we have a delocated wheel with a GEOS dylib.
-        geos_whl_dylib = os.path.abspath(os.path.join(os.path.dirname(
-            __file__), '.dylibs/libgeos_c.1.dylib'))
-        if os.path.exists(geos_whl_dylib):
-            _lgeos = CDLL(geos_whl_dylib)
-            LOG.debug("Found GEOS DLL: %r, using it.", _lgeos)
-
-        else:
-            if hasattr(sys, 'frozen'):
-                try:
-                    # .app file from py2app
-                    alt_paths = [os.path.join(os.environ['RESOURCEPATH'],
-                                '..', 'Frameworks', 'libgeos_c.dylib')]
-                except KeyError:
-                    # binary from pyinstaller
-                    alt_paths = [
-                        os.path.join(sys.executable, 'libgeos_c.dylib')]
-            else:
-                alt_paths = [
-                    # The Framework build from Kyng Chaos
-                    "/Library/Frameworks/GEOS.framework/Versions/Current/GEOS",
-                    # macports
-                    '/opt/local/lib/libgeos_c.dylib',
-                ]
-            _lgeos = load_dll('geos_c', fallbacks=alt_paths)
+        _lgeos = load_dll('geos_c', fallbacks=alt_paths)
 
     free = load_dll('c').free
     free.argtypes = [c_void_p]
     free.restype = None
 
 elif sys.platform == 'win32':
-    try:
-        egg_dlls = os.path.abspath(os.path.join(os.path.dirname(__file__),
-                                   "DLLs"))
-        wininst_dlls = os.path.abspath(os.__file__ + "../../../DLLs")
-        original_path = os.environ['PATH']
-        os.environ['PATH'] = "%s;%s;%s" % \
-            (egg_dlls, wininst_dlls, original_path)
-        _lgeos = CDLL("geos.dll")
-    except (ImportError, WindowsError, OSError):
-        raise
-
-    def free(m):
+    if os.getenv('CONDA_PREFIX', ''):
+        # conda package.
+        _lgeos = CDLL(os.path.join(sys.prefix, 'Library', 'bin', 'geos_c.dll'))
+    else:
         try:
-            cdll.msvcrt.free(m)
-        except WindowsError:
-            # XXX: See http://trac.gispython.org/projects/PCL/ticket/149
-            pass
+            egg_dlls = os.path.abspath(
+                os.path.join(os.path.dirname(__file__), 'DLLs'))
+            if hasattr(sys, '_MEIPASS'):
+                wininst_dlls = sys._MEIPASS
+            elif hasattr(sys, "frozen"):
+                wininst_dlls = os.path.normpath(
+                    os.path.abspath(sys.executable + '../../DLLS'))
+            else:
+                wininst_dlls = os.path.abspath(os.__file__ + "../../../DLLs")
+            original_path = os.environ['PATH']
+            os.environ['PATH'] = "%s;%s;%s" % \
+                (egg_dlls, wininst_dlls, original_path)
+            _lgeos = load_dll("geos_c.dll")
+        except (ImportError, WindowsError, OSError):
+            raise
+
+        def free(m):
+            try:
+                cdll.msvcrt.free(m)
+            except WindowsError:
+                # XXX: See http://trac.gispython.org/projects/PCL/ticket/149
+                pass
 
 elif sys.platform == 'sunos5':
     _lgeos = load_dll('geos_c', fallbacks=['libgeos_c.so.1', 'libgeos_c.so'])
@@ -159,11 +181,9 @@ else:  # other *nix systems
 
 
 def _geos_version():
-    # extern const char GEOS_DLL *GEOSversion();
     GEOSversion = _lgeos.GEOSversion
     GEOSversion.restype = c_char_p
     GEOSversion.argtypes = []
-    #define GEOS_CAPI_VERSION "@VERSION@-CAPI-@CAPI_VERSION@"
     geos_version_string = GEOSversion()
     if sys.version_info[0] >= 3:
         geos_version_string = geos_version_string.decode('ascii')
@@ -211,51 +231,26 @@ if geos_version >= (3, 1, 0):
         [EXCEPTION_HANDLER_FUNCTYPE, EXCEPTION_HANDLER_FUNCTYPE]
     _lgeos.finishGEOS_r.argtypes = [c_void_p]
 
-# Exceptions
 
+def make_logging_callback(func):
+    """Error or notice handler callback producr
 
-class ReadingError(Exception):
-    pass
-
-
-class DimensionError(Exception):
-    pass
-
-
-class TopologicalError(Exception):
-    pass
-
-
-class PredicateError(Exception):
-    pass
-
-# While this function can take any number of positional arguments when
-# called from Python and GEOS expects its error handler to accept any
-# number of arguments (like printf), I'm unable to get ctypes to make
-# a callback object from this function that will accept any number of
-# arguments.
-#
-# At the moment, functions in the GEOS C API only pass 0 or 1 arguments
-# to the error handler. We can deal with this, but when if that changes,
-# Shapely may break.
-
-def handler(level):
-    def callback(fmt, *args):
+    Wraps a logger method, func, as a GEOS callback.
+    """
+    def callback(fmt, *fmt_args):
         fmt = fmt.decode('ascii')
         conversions = re.findall(r'%.', fmt)
-        log_vals = []
-        for spec, arg in zip(conversions, args):
-            if spec == '%s' and arg is not None:
-                log_vals.append(string_at(arg).decode('ascii'))
-            else:
-                LOG.error("An error occurred, but the format string "
-                          "'%s' could not be converted.", fmt)
-                return
-        getattr(LOG, level)(fmt, *log_vals)
+        args = [
+            string_at(arg).decode('ascii')
+            for spec, arg in zip(conversions, fmt_args)
+            if spec == '%s' and arg is not None]
+
+        func(fmt, *args)
+
     return callback
 
-error_handler = handler('error')
-notice_handler = handler('warning')
+error_handler = make_logging_callback(LOG.error)
+notice_handler = make_logging_callback(LOG.info)
 
 error_h = EXCEPTION_HANDLER_FUNCTYPE(error_handler)
 notice_h = EXCEPTION_HANDLER_FUNCTYPE(notice_handler)
@@ -280,12 +275,16 @@ class WKTReader(object):
 
     def read(self, text):
         """Returns geometry from WKT"""
+        if not isinstance(text, text_types):
+            raise TypeError("Only str is accepted.")
         if sys.version_info[0] >= 3:
-            text = text.encode('ascii')
-        geom = self._lgeos.GEOSWKTReader_read(self._reader, c_char_p(text))
+            text = text.encode()
+        c_string = c_char_p(text)
+        geom = self._lgeos.GEOSWKTReader_read(self._reader, c_string)
         if not geom:
-            raise ReadingError("Could not create geometry because of errors "
-                               "while reading input.")
+            raise WKTReadingError(
+                "Could not create geometry because of errors "
+                "while reading input.")
         # avoid circular import dependency
         from shapely.geometry.base import geom_factory
         return geom_factory(geom)
@@ -424,8 +423,9 @@ class WKBReader(object):
         geom = self._lgeos.GEOSWKBReader_read(
             self._reader, c_char_p(data), c_size_t(len(data)))
         if not geom:
-            raise ReadingError("Could not create geometry because of errors "
-                               "while reading input.")
+            raise WKBReadingError(
+                "Could not create geometry because of errors "
+                "while reading input.")
         # avoid circular import dependency
         from shapely import geometry
         return geometry.base.geom_factory(geom)
@@ -437,8 +437,9 @@ class WKBReader(object):
         geom = self._lgeos.GEOSWKBReader_readHEX(
             self._reader, c_char_p(data), c_size_t(len(data)))
         if not geom:
-            raise ReadingError("Could not create geometry because of errors "
-                               "while reading input.")
+            raise WKBReadingError(
+                "Could not create geometry because of errors "
+                "while reading input.")
         # avoid circular import dependency
         from shapely import geometry
         return geometry.base.geom_factory(geom)
@@ -559,13 +560,20 @@ def errcheck_just_free(result, func, argtuple):
     else:
         return retval
 
+
 def errcheck_null_exception(result, func, argtuple):
-    """Wraps errcheck_just_free, raising a TopologicalError if result is NULL"""
+    """Wraps errcheck_just_free
+
+    Raises TopologicalError if result is NULL.
+    """
     if not result:
-        raise TopologicalError("The operation '{0}' could not be performed."
-            "Likely cause is invalidity of the geometry.".format(func.__name__))
+        raise TopologicalError(
+            "The operation '{}' could not be performed."
+            "Likely cause is invalidity of the geometry.".format(
+                func.__name__))
     return errcheck_just_free(result, func, argtuple)
 
+
 def errcheck_predicate(result, func, argtuple):
     """Result is 2 on exception, 1 on True, 0 on False"""
     if result == 2:
@@ -674,7 +682,7 @@ class LGEOS310(LGEOSBase):
         for key in [x for x in keys if not x.endswith('_r')]:
             if key + '_r' in keys:
                 reentr_func = getattr(self._lgeos, key + '_r')
-                attr = ftools.partial(reentr_func, self.geos_handle)
+                attr = partial(reentr_func, self.geos_handle)
                 attr.__name__ = reentr_func.__name__
                 setattr(self, key, attr)
             else:
@@ -782,12 +790,16 @@ class LGEOS320(LGEOS311):
         super(LGEOS320, self).__init__(dll)
 
         if geos_version >= (3, 2, 0):
-            def parallel_offset(geom, distance, resolution=16, join_style=1, mitre_limit=5.0, side='right'):
+
+            def parallel_offset(geom, distance, resolution=16, join_style=1,
+                                mitre_limit=5.0, side='right'):
                 side = side == 'left'
                 if distance < 0:
                     distance = abs(distance)
                     side = not side
-                return self.GEOSSingleSidedBuffer(geom, distance, resolution, join_style, mitre_limit, side)
+                return self.GEOSSingleSidedBuffer(
+                    geom, distance, resolution, join_style, mitre_limit, side)
+
             self.methods['parallel_offset'] = parallel_offset
 
         self.methods['project'] = self.GEOSProject
@@ -796,6 +808,7 @@ class LGEOS320(LGEOS311):
         self.methods['interpolate_normalized'] = \
             self.GEOSInterpolateNormalized
         self.methods['buffer_with_style'] = self.GEOSBufferWithStyle
+        self.methods['hausdorff_distance'] = self.GEOSHausdorffDistance
 
 
 class LGEOS330(LGEOS320):
@@ -811,23 +824,27 @@ class LGEOS330(LGEOS320):
         # GEOSPolygonize_full. We patch it in explicitly here.
         key = 'GEOSPolygonize_full'
         func = getattr(self._lgeos, key + '_r')
-        attr = ftools.partial(func, self.geos_handle)
+        attr = partial(func, self.geos_handle)
         attr.__name__ = func.__name__
         setattr(self, key, attr)
 
         for pred in (self.GEOSisClosed,):
             pred.func.errcheck = errcheck_predicate
 
-        def parallel_offset(geom, distance, resolution=16, join_style=1, mitre_limit=5.0, side='right'):
+        def parallel_offset(geom, distance, resolution=16, join_style=1,
+                            mitre_limit=5.0, side='right'):
             if side == 'right':
                 distance *= -1
-            return self.GEOSOffsetCurve(geom, distance, resolution, join_style, mitre_limit)
-        self.methods['parallel_offset'] = parallel_offset
+            return self.GEOSOffsetCurve(
+                geom, distance, resolution, join_style, mitre_limit)
 
+        self.methods['parallel_offset'] = parallel_offset
         self.methods['unary_union'] = self.GEOSUnaryUnion
         self.methods['is_closed'] = self.GEOSisClosed
         self.methods['cascaded_union'] = self.methods['unary_union']
         self.methods['snap'] = self.GEOSSnap
+        self.methods['shared_paths'] = self.GEOSSharedPaths
+        self.methods['buffer_with_params'] = self.GEOSBufferWithParams
 
 
 class LGEOS340(LGEOS330):
@@ -842,7 +859,18 @@ class LGEOS340(LGEOS330):
         self.methods['nearest_points'] = self.GEOSNearestPoints
 
 
-if geos_version >= (3, 4, 0):
+class LGEOS350(LGEOS340):
+    """Proxy for GEOS 3.5.0-CAPI-1.9.0
+    """
+
+    def __init__(self, dll):
+        super(LGEOS350, self).__init__(dll)
+        self.methods['clip_by_rect'] = self.GEOSClipByRect
+
+
+if geos_version >= (3, 5, 0):
+    L = LGEOS350
+elif geos_version >= (3, 4, 0):
     L = LGEOS340
 elif geos_version >= (3, 3, 0):
     L = LGEOS330
